/* libStreamedXML/src/lib/Core/Callback.h
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lsx {



/*! \brief This object is the callback used in lsx::Parser.

The functions in this callback may be overridden to handle "events"
encountered when parsing Streamed XML. For instance, hitting an open
tag will generate an element() event. Only well-formed XML will ever
get through to the callback object.

Default implementations are provided for all the functions. These
implementations do nothing, so you are free to implement only those
functions that you need.

You may throw any exceptions you like from the callback handlers; these
will filter through the parser layer and reach the character decoder
layer.

*/
class Callback {
public:
    /// Destructor (does nothing).
    virtual ~Callback()
    { }



    /*! \brief Stream restart marker.

    \param data Data held in the marker (may be empty).

    This function is called whenever a stream restart marker is
    encountered. This means that, if you are keeping track of the parsed
    data in any way, you should now reset that state (e.g. you should
    clear a stack of which element you are currently in).

    */
    virtual void streamRestart(const std::wstring& data)
    {
        (void)data;
    }



    /*! \brief Comment.

    \param data Data held in the comment (may be empty).

    This function is called whenever a comment is encountered. It is
    mainly used if you want to transform the input in some ways, but
    leave the comments intact.

    */
    virtual void comment(const std::wstring& data)
    {
        (void)data;
    }



    /*! \brief Processing instruction.

    \param target The processing instruction's target.
    \param data Any data following the target (may be empty).

    This function is called whenever a PI is encountered. It includes
    both the target and any data following that target.

    */
    virtual void PI(const std::wstring& target, const std::wstring& data)
    {
        (void)target;
        (void)data;
    }



    /*! \brief Element start tag.

    \param elemName The element's name.
    \param elemAttrs Any attributes the element may have (may be empty).

    This is called whenever an element start tag is encountered. It is
    called with the element's name and attributes. If the element tag is
    an empty tag, then closeTag() will be called immediately afterward.

    */
    virtual void element(const std::wstring& elemName,
                         const std::map<std::wstring, std::wstring>& elemAttrs)
    {
        (void)elemName;
        (void)elemAttrs;
    }



    /*! \brief Element close tag.

    \param elemName The element's name.

    This is called whenever a close tag is encountered. It is also
    generated after the element() event if an empty element tag is
    encountered.

    */
    virtual void closeTag(const std::wstring& elemName)
    {
        (void)elemName;
    }



    /*! \brief Ignorable whitespace.

    \param ws The actual whitespace encountered.

    This is called whenever whitespace is encountered between element
    tags, but there is no actual content. It is included simply because
    you may wish to transform the input in some ways, but keep the
    whitespace the same.

    If an element does contain content but is surrounded by whitespace,
    this function will not be called: the whitespace will be included
    in the reported content.

    */
    virtual void whiteSpace(const std::wstring& ws)
    {
        (void)ws;
    }



    /*! \brief Element content.

    \param data The element's content.

    This function is called whenever non-whitespace content is
    encountered within an element. Any whitespace within the content (or
    at the start or end) is reported verbatim.

    */
    virtual void content(const std::wstring& data)
    {
        (void)data;
    }



    /*! \brief Character data.

    \param data The character data.

    This is used to report a chunk of character data enclosed in a CDATA
    marked section. This means it might contain the text "&amp;", but
    this shouldn't be expanded as an entity, so there would be no way of
    disambiguating it from an entity reference if \a expandEntities is
    \a false.

    The default behaviour is to pass this straight to content().

    */
    virtual void cdata(const std::wstring& data)
    {
        content(data);
    }



    /*! \brief Entity reference.

    \param name The name of the entity being referred to.
    \param[out] value The expanded text.
    \retval true if the entity is valid.
    \retval false if the entity is not valid.

    This function is called whenever an entity reference is encountered,
    and \a expandEntities is \a true in your parser. The five standard
    entities will be expanded automatically, as will character entities,
    but anything else will be passed to this function.

    */
    virtual bool entityRef(const std::wstring& name, std::wstring& value)
    {
        (void)name;
        (void)value;
        return false;
    }
};



};
/* options for text editors
kate: replace-trailing-space-save true; space-indent true; tab-width 4;
vim: expandtab:ts=4:sw=4
*/
