/* libStreamedXML/src/lib/Core/Parser.h
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

namespace lsx {



/*! \brief Streamed XML parser object.

This object sits between the callback layer (see lsx::Callback) and the
character decoding layer. It is a finite state machine which parses
character data as input and gives Streamed XML events as output. The
FSM design gives both speed and robustness. FSMs are good for dealing
with true data streams since their behaviour can easily be controlled to
avoid becoming unstable, and also because it is easy to provide "reset"
functionality (as in the case of a stream restart marker).

Data must be fed into the parser through either of the feedChar() or
feedCharData() functions.
Any stream restarts detected by the character decoding layer can be
signalled through the streamRestart() function. The parser will accept
stream reset markers, but only if they occur at stream level; if the
data stream became corrupted (e.g. the end of a comment tag was missed),
the Parser class would not be looking for stream restart markers and
would thus ignore them.

A parser is tied to a single lsx::Callback object, which cannot be
changed. You could potentially get around this by providing a
multiplexing Callback object. As many streams as you like may provide
input to a Parser, but this would only make sense if you synchronised
their inputs in some manner. Similarly, as many Parser objects as you
like may share the same callback, but again some form of synchronisation
would be needed.

Should an error occur in processing the XML (i.e. it is not well
formed), a NotWellFormed exception will be thrown. It is also possible
for a callback function to throw any exception, which will propagate
through to the character decoding layer. If this occurs, the character
decoding layer should switch into a state where it searches for a
stream restart marker, and not pass any more data to the Parser until it
encounters one. At this point, the streamRestart() function should be
called and processing can continue.

*/
class Parser {
private:
    //BEGIN // Enumerations for finite state machine ///////////////////
    enum State {
        StateNone, // at element or stream level
        StateRestartMarker, // after <![RESTART[
        StateRestartMarker1, // after first ]
        StateRestartMarker2, // after second ]
        StateOpen, // after <
        StateOpenBang, // after <!
        StateOpenCdataMarker, // after <![
        StateCDATA,
        StateCDATA1, // first ]
        StateCDATA2, // second ]
        StateData, // like StateNone, but don't skip whitespace
        StateOpenComment, // after <!-
        StateComment,
        StateComment2,
        StateComment3,
        StatePI, // after <?    (we are currently parsing the target)
        StatePIData, // after target
        StatePI2, // first ? encountered (in target)
        StatePI3, // first ? encountered (in data)
        StateClose, // "</" encountered
        StateClosing, // "</x"
        StateNeedClose2, // "</xyz "
        StateElemName, // <x encountered
        StateElemTag,  // first space after "<name" encountered
        StateElemAttrName,
        StateElemAttrEq, // name=
        StateElemAttrVal, // name=' or name="
        StateElemAttrDone,  // name='value'
        StateNeedClose, // <elem/
        StateEntity, // &
        StateCharEntity, // &#
        StateEntityName, // &x
        StateError
    };



    enum TokenClass {
        ClassEntity,
        ClassNameChar,
        ClassNameStartChar,
        ClassOpenTag,
        ClassWhitespace,
        ClassOther
    };
    //END // Enumerations //////////////////////////////////////////////

    static const wchar_t* xmlCdataMarker;
    static const wchar_t* xmlRestartMarker;
    
    // the callback object to use
    Callback* callback;
    
    // entity parsing
    std::wstring entityRef(const std::wstring& ent);
    bool expandEntities;
    bool parsingAttr;

    // used for parsing elements
    std::vector<std::wstring> elemStack;
    std::wstring elemName;
    std::map<std::wstring, std::wstring> elemAttrs;

    // parsing state
    State state;
    std::wstring buffer, buffer2, buffer3;
    wchar_t entityChar;
    bool singleQuote;
    bool skipNextNewline;
    int elementDepth, xmlCount, restartCount;

    // for reporting errors
    int line, col;

public:
    /*! \brief Constructor.

    \param callback The callback object to use.
    \param expandEntities \a true if you want entities to be expanded,
        \a false to keep them inline.

    This sets up the finite state machine and records the callback
    object to use. It doesn't generate any events.

    */
    Parser(Callback* callback, bool expandEntities);



    /*! \brief Feed parser a character.

    \param ch The character to parse.
    \throws NotWellFormed If the Streamed XML is not well formed.
    \throws (anything) Anything from the callback object.

    This function will cause the finite state machine to process an
    input character. It may generate some callback events. It will
    throw a NotWellFormed exception should it encounter any Streamed XML
    structure that is not well formed; it will also propagate any
    exceptions thrown from the higher callback layer.

    */
    void feedChar(wchar_t ch);



    /*! \brief Feed parser some character data.

    \param data A pointer to the buffer of data.
    \param amount The number of characters to parse (may be zero).
    \throws NotWellFormed If the Streamed XML is not well formed.
    \throws (anything) Anything from the callback object.

    This function will cause the finite state machine to process some
    input characters. It may generate some callback events. It will
    throw a NotWellFormed exception should it encounter any Streamed XML
    structure that is not well formed; it will also propagate any
    exceptions thrown from the higher callback layer.

    */
    inline void feedCharData(const wchar_t* data, size_t amount)
    {
        for(size_t pos = 0; pos < amount; ++pos) feedChar(data[pos]);
    }



    /*! \brief Feed parser some character data.

    \param dataStr The data, as a string.
    \throws NotWellFormed If the Streamed XML is not well formed.
    \throws (anything) Anything from the callback object.

    This function will cause the finite state machine to process some
    input characters. It may generate some callback events. It will
    throw a NotWellFormed exception should it encounter any Streamed XML
    structure that is not well formed; it will also propagate any
    exceptions thrown from the higher callback layer.

    */
    inline void feedCharData(const std::wstring& dataStr)
    {
        std::wstring::size_type pos = 0, len = dataStr.size();
        for(; pos < len; ++pos) feedChar(dataStr[pos]);
    }

    
    
    /*! \brief Reset the stream.
    
    This function should be called to reset the stream. This could be due to e.g. a new file or a
    restart marker.
    
     */
    void reset();



    /*! \brief Restart marker detected.

    This function should be called when a stream restart marker (
    \c &lt;![RESTART[ ) has been detected. The next character of
    data to be passed to the parser should be the first character of
    restart marker data (or a \c ']', if there is no data).

    This function will cause the parser to generate a streamRestart()
    callback.

    */
    void streamRestart();
};



}
/* options for text editors
kate: replace-trailing-space-save true; space-indent true; tab-width 4;
vim: expandtab:ts=4:sw=4
*/
