/* libStreamedXML/src/tests/Structure.cpp
 *
 *  (c)2005, Laurence Withers. Released under the GNU GPL. See file
 *  COPYING for more information / terms of license.
*/

#include "StreamedXML"

#include <iostream>
#include <list>



class TestFailed { };



//BEGIN // Test results ////////////////////////////////////////////////



#define DYN_CAST_CHECK(_type, _name) \
    TestResult* r = getExpected( _name ); \
    _type * rr = dynamic_cast< _type *>(r); \
    if(!rr) { \
        std::wcerr << L"Expected " << r->name() << L", got " _name L".\n"; \
        delete r; \
        throw TestFailed(); \
    }



class TestResult {
public:
    virtual ~TestResult() { }
    virtual std::wstring name() const = 0;
};

class TestResultException : public TestResult {
public:
    virtual std::wstring name() const { return L"exception"; }
};

class TestResultStreamRestart : public TestResult {
public:
    std::wstring data;
    TestResultStreamRestart(const std::wstring& data) : data(data) { }
    bool verify(const std::wstring& data) const { return data == this->data; }
    virtual std::wstring name() const { return L"stream restart marker"; }
};

class TestResultComment : public TestResult {
public:
    std::wstring data;
    TestResultComment(const std::wstring& data) : data(data) { }
    virtual std::wstring name() const { return L"comment"; }

    void verify(const std::wstring& data) const
    {
        if(data != this->data) {
            std::wcerr << name() << L": expecting ``" << this->data
                  << L"'', got ``" << data << L"''.\n";
            throw TestFailed();
        }
    }
};

class TestResultCloseTag : public TestResult {
public:
    std::wstring data;
    TestResultCloseTag(const std::wstring& data) : data(data) { }
    virtual std::wstring name() const { return L"close tag"; }

    void verify(const std::wstring& data) const
    {
        if(data != this->data) {
            std::wcerr << name() << L": expecting ``" << this->data
                  << L"'', got ``" << data << L"''.\n";
            throw TestFailed();
        }
    }
};

class TestResultContent : public TestResult {
public:
    std::wstring data;
    TestResultContent(const std::wstring& data) : data(data) { }
    virtual std::wstring name() const { return L"content"; }

    void verify(const std::wstring& data) const
    {
        if(data != this->data) {
            std::wcerr << name() << L": expecting ``" << this->data
                  << L"'', got ``" << data << L"''.\n";
            throw TestFailed();
        }
    }
};

class TestResultWhiteSpace : public TestResult {
public:
    std::wstring data;
    TestResultWhiteSpace(const std::wstring& data) : data(data) { }
    virtual std::wstring name() const { return L"whitespace"; }

    void verify(const std::wstring& data) const
    {
        if(data != this->data) {
            std::wcerr << name() << L": expecting ``" << this->data
                  << L"'', got ``" << data << L"''.\n";
            throw TestFailed();
        }
    }
};

class TestResultCDATA : public TestResult {
public:
    std::wstring data;
    TestResultCDATA(const std::wstring& data) : data(data) { }
    virtual std::wstring name() const { return L"CDATA marker"; }

    void verify(const std::wstring& data) const
    {
        if(data != this->data) {
            std::wcerr << name() << L": expecting ``" << this->data
                  << L"'', got ``" << data << L"''.\n";
            throw TestFailed();
        }
    }
};

class TestResultEntityRef : public TestResult {
public:
    std::wstring data;
    TestResultEntityRef(const std::wstring& data) : data(data) { }
    virtual std::wstring name() const { return L"entity reference"; }

    void verify(const std::wstring& data) const
    {
        if(data != this->data) {
            std::wcerr << name() << L": expecting ``" << this->data
                  << L"'', got ``" << data << L"''.\n";
            throw TestFailed();
        }
    }
};

class TestResultPI : public TestResult {
public:
    std::wstring target, data;
    TestResultPI(const std::wstring& target, const std::wstring& data)
        : target(target), data(data) { }
    virtual std::wstring name() const { return L"processing instruction"; }

    void verify(const std::wstring& target, const std::wstring& data) const
    {
        if(target != this->target) {
            std::wcerr << name() << L": expecting ``" << this->target
                  << L"'', got ``" << target << L"''.\n";
            throw TestFailed();
        }
        if(data != this->data) {
            std::wcerr << name() << L": expecting ``" << this->data
                  << L"'', got ``" << data << L"''.\n";
            throw TestFailed();
        }
    }
};

class TestResultElement : public TestResult {
public:
    std::wstring ename;
    std::map<std::wstring, std::wstring> attrs;
    TestResultElement(const std::wstring& ename) : ename(ename) { }
    virtual std::wstring name() const { return L"open tag"; }

    void verify(const std::wstring& xname, const std::map<std::wstring, std::wstring>& xattrs)
    {
        if(ename != xname) {
            std::wcerr << name() << L": expecting ``" << ename
                  << L"'', got ``" << xname << L"''.\n";
            throw TestFailed();
        }
        std::map<std::wstring, std::wstring>::const_iterator i1, i2, end1, end2;
        i1 = attrs.begin(); end1 = attrs.end();
        i2 = xattrs.begin(); end2 = xattrs.end();

        for(; (i1 != end1) && (i2 != end2); ++i1, ++i2) {
            if(i1->first != i2->first) {
                std::wcerr << name() << L": attribute name: expecting ``" << i1->first
                      << L"'', got ``" << i2->first << L"''.\n";
                throw TestFailed();
            }
            if(i1->second != i2->second) {
                std::wcerr << name() << L": attribute value: expecting ``" << i1->second
                      << L"'', got ``" << i2->second << L"''.\n";
                throw TestFailed();
            }
        }

        if((i1 != end1) || (i2 != end2)) {
            std::wcerr << name() << L": attributes don't match.\n";
            throw TestFailed();
        }
    }
};



//END // Test results //////////////////////////////////////////////////
//BEGIN // Test parser /////////////////////////////////////////////////



class TestParser : private lsx::Callback {
private:
    std::list<TestResult*> expected;
    TestResult* getExpected(const std::wstring& dataType)
    {
        if(expected.empty()) {
            std::wcerr << L"Further data (" << dataType
                  << L") encountered after end of test.\n";
            throw TestFailed();
        }
        TestResult* r = expected.front();
        expected.pop_front();
        return r;
    }
    
    virtual void streamRestart(const std::wstring& data)
    {
        DYN_CAST_CHECK(TestResultStreamRestart, L"stream restart marker")
                rr->verify(data);
    }

    virtual void comment(const std::wstring& data)
    {
        DYN_CAST_CHECK(TestResultComment, L"comment")
                rr->verify(data);
    }

    virtual void closeTag(const std::wstring& data)
    {
        DYN_CAST_CHECK(TestResultCloseTag, L"close tag")
                rr->verify(data);
    }

    virtual void whiteSpace(const std::wstring& data)
    {
        DYN_CAST_CHECK(TestResultWhiteSpace, L"whitespace")
                rr->verify(data);
    }

    virtual void content(const std::wstring& data)
    {
        DYN_CAST_CHECK(TestResultContent, L"Content")
                rr->verify(data);
    }

    virtual void cdata(const std::wstring& data)
    {
        DYN_CAST_CHECK(TestResultCDATA, L"CDATA marker")
                rr->verify(data);
    }

    virtual std::wstring entityRef(const std::wstring& data)
    {
        DYN_CAST_CHECK(TestResultEntityRef, L"entity reference")
                rr->verify(data);
        return data;
    }

    virtual void PI(const std::wstring& target, const std::wstring& data)
    {
        DYN_CAST_CHECK(TestResultPI, L"processing instruction")
                rr->verify(target, data);
    }

    virtual void element(const std::wstring& name, const std::map<std::wstring, std::wstring>& attrs)
    {
        DYN_CAST_CHECK(TestResultElement, L"open tag")
                rr->verify(name, attrs);
    }
    
    // XML parser objects
    lsx::Parser xmlParser;
    lsx::Decoder xmlDecoder;

public:
    TestParser()
        : xmlParser(this, true), xmlDecoder(&xmlParser)
    {
    }
    
    virtual ~TestParser()
    {
        while(!expected.empty()) {
            delete expected.front();
            expected.pop_front();
        }
    }

    void finish()
    {
        if(!expected.empty()) {
            std::wcerr << L"Expecting some more data ("
                    << expected.front()->name()
                    << L").\n";
            throw TestFailed();
        }
    }

    void expect(TestResult* r)
    {
        expected.push_back(r);
    }

    virtual void feedData(const char* bytes, size_t amt)
    {
        while(amt) {
            try {
                xmlDecoder.feedData(bytes, std::min(amt, size_t(6)));
            }
            catch(lsx::Exception& e) {
                DYN_CAST_CHECK(TestResultException, L"exception")
                delete r;
            }
            amt -= std::min(amt, size_t(6));
            bytes += 6;
        }
    }
};



//END // Test parser ///////////////////////////////////////////////////
//BEGIN // Tests ///////////////////////////////////////////////////////



void test1(TestParser* p)
{
    std::wcout << L"Test 1: " << std::flush;

    p->expect(new TestResultPI(L"targ", L" data  data "));
    p->expect(new TestResultWhiteSpace(L"\n"));
    p->expect(new TestResultComment(L" comment "));
    p->expect(new TestResultWhiteSpace(L"\n"));
    p->expect(new TestResultElement(L"elemOne"));
    p->expect(new TestResultWhiteSpace(L" "));
    p->expect(new TestResultContent(L"with added content\n"));
    p->expect(new TestResultException());
    p->expect(new TestResultStreamRestart(L"data]>]]"));
    p->expect(new TestResultWhiteSpace(L"\n"));
    p->expect(new TestResultElement(L"elemTwo"));
    p->expect(new TestResultWhiteSpace(L"\n"));
    p->expect(new TestResultCloseTag(L"elemTwo"));

    std::string td = "<?targ  data  data ?>\n"
                "<!-- comment -->\n"
                "<elemOne> with added content\n"
                "<![RESTART[data]>]]]]>\n"
                "<elemTwo>\n"
                "</elemTwo>\n";
    p->feedData(td.data(), td.size());
    p->finish();

    std::wcout << L"OK.\n";
}



void test2(TestParser* p)
{
    std::wcout << L"Test 2: " << std::flush;

    p->expect(new TestResultElement(L"reading"));
    p->expect(new TestResultElement(L"type"));
    p->expect(new TestResultContent(L"integer"));
    p->expect(new TestResultCloseTag(L"type"));
    p->expect(new TestResultElement(L"value"));
    p->expect(new TestResultContent(L"0"));
    p->expect(new TestResultCloseTag(L"value"));
    p->expect(new TestResultCloseTag(L"reading"));
    p->expect(new TestResultElement(L"reading"));
    p->expect(new TestResultElement(L"type"));
    p->expect(new TestResultContent(L"integer"));
    p->expect(new TestResultCloseTag(L"type"));
    p->expect(new TestResultElement(L"value"));
    p->expect(new TestResultContent(L"1"));
    p->expect(new TestResultCloseTag(L"value"));
    p->expect(new TestResultCloseTag(L"reading"));

    std::string td = "<reading><type>integer</type><value>0</value></reading>"
                "<reading><type>integer</type><value>1</value></reading>";
    p->feedData(td.data(), td.size());
    p->finish();

    std::wcout << L"OK.\n";
}



void test3(TestParser* p)
{
    std::wcout << L"Test 3: " << std::flush;

    TestResultElement* el = new TestResultElement(L"elemName");
    el->attrs[L"attr1"] = L"\"value1\"";
    el->attrs[L"attr2"] = L"'value2'";
    el->attrs[L"attr3"] = L"\"<&>'";
    p->expect(el);

    p->expect(new TestResultContent(L"<"));
    p->expect(new TestResultContent(L"&"));
    p->expect(new TestResultContent(L">"));
    p->expect(new TestResultEntityRef(L"myEntity"));
    p->expect(new TestResultContent(L"myEntity\n    "));
    p->expect(new TestResultCDATA(L"<&>]]]]"));
    p->expect(new TestResultWhiteSpace(L"\n"));
    p->expect(new TestResultCloseTag(L"elemName"));

    std::string td = "<elemName attr1='\"value1\"' attr2=\"'value2'\"\n"
                "          attr3='&quot;&lt;&amp;&gt;&apos;'\n"
                "    >&lt;&amp;&gt;&myEntity;\n"
                "    <![CDATA[<&>]]]]]]>\n"
                "</elemName >\n";
    p->feedData(td.data(), td.size());
    p->finish();

    std::wcout << L"OK.\n";
}



//END // Tests /////////////////////////////////////////////////////////



int main(int argc, char* argv[])
{
    if(argc == 2 && !strcmp(argv[1], "--print-summary")) {
        std::wcout << L"XML structure regression test.\n";
        return 0;
    }

    if(argc != 1) {
        std::wcerr << L"No arguments expected.\n";
        return 1;
    }

    int ret = 0;
    TestParser* p = 0;
    
#define TEST_n(_n) \
    try { \
        std::wcout << "======== Test " #_n " ========" << std::endl; \
        p = new TestParser(); \
        test ## _n (p); \
        delete p; \
        p = 0; \
    } \
    catch(TestFailed&) { \
        std::wcerr << "Test failed." << std::endl; \
        ret = 1; \
    } \
    catch(std::exception& e) { \
        std::cerr << e.what() << std::endl; \
        ret = 1; \
    } \
    std::wcout << std::endl;
    
    TEST_n(1);
    TEST_n(2);
    TEST_n(3);
        
    delete p;
    return ret;
}
